/*
Copyright (C) 2006 Fred Eaker <fredeaker@gmail.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/
package net.sourceforge.vyasa;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.charset.Charset;

import monq.stuff.EncodingDetector;

import org.apache.commons.lang.SystemUtils;
import org.mozilla.intl.chardet.nsDetector;
import org.mozilla.intl.chardet.nsICharsetDetectionObserver;
import org.nuxeo.runtime.util.FileUtils;

import com.glaforge.i18n.io.CharsetToolkit;
import com.ibm.icu.text.CharsetDetector;
import com.ibm.icu.text.CharsetMatch;
import com.sun.syndication.io.XmlReader;

import cpdetector.io.ParsingDetector;

/**
 * Tests a series of files against character set detection libraries and
 * produces  results.
 * @author Fred Eaker <fredeaker@gmail.com>
 */
public class CharsetDetectorTester implements nsICharsetDetectionObserver
{
	/* The testPath and outputFilename variables are used to stored arguments
	 * passed to the main method.
	 */
	
	private static String testPath;
	private static String outputFilename;
	
	/* This static class is given an instance variable to facilitate the
	 * nsICharsetDetectionObserver. This allows the class to be passed to itself.
	 * See the Notify(String string) method.
	 */
	
	private static CharsetDetectorTester self;
	private static String jchardetResult;
	
	private static FileWriter outputFileWriter;
	
	/**
	 * The prviate constructor is only used within the class itself. Its purpose
	 * is to generate a static instance that can be used to pass the class to
	 * itself. (See the Notify(String string) method)
	 */
	private CharsetDetectorTester() {self = this;}
	
	/**
	 * Uses the reults of command line argument processing to generate an array
	 * of Files to process. Each file is then passed to each character set
	 * detector and the results are written to an output file.
	 * @param args path to test files and name of output file
	 */
	public static void main(String[] args)
	{
		processArguments(args);
		tryOutputFileWriter();
		
		/* Initalize the self variable. */
		
		new CharsetDetectorTester();
		
		/* Generate an array of files to process. */
		
		File[] testFiles = new File(testPath).listFiles();
		
		for (int i = 0; i < testFiles.length; i++)
		{
			/* Since the output file will be saved in the same path as the test
			 * files, avoid processing it. */
			
			if ((!testFiles[i].getName().equals(outputFilename)) 
				&& (!testFiles.clone()[i].isDirectory()))
			{
				System.out.println("Detecting: " + testFiles[i].getName());
				detectCharset(testFiles[i]);
			}
		}
		try {outputFileWriter.close();} catch (IOException e) {e.printStackTrace();}
	}

	/**
	 * Writes the results of character set dection to the output file by passing
	 * the file parameter to a series of detectors.  
	 * @param file - the file for character set detection
	 */
	private static void detectCharset(File file)
	{
		try
		{
			outputFileWriter.write(
					file.getName() + "|" +
					getJChardet(file) + "|" +
					getCpDetect(file) + "|" + 
					getICU4J(file) + "|" +
					getMonq(file) + "|" +
					getRome(file) + "|" +
					getGlaforge(file) + "\n");
		}
		catch (IOException e) {e.printStackTrace();}
	}
	
	/**
	 * Returns the results of the {@see org.mozilla.intl.chardet.nsDetector}
	 * library.
	 * @param file - the file for character set detection
	 * @return String results of character set detection
	 */
	private static String getJChardet(File file)
	{
		try
		{
			jchardetResult = null;
			nsDetector jchardetector = new nsDetector(0);
			jchardetector.Init(self);
			
			BufferedInputStream bis =
				new BufferedInputStream(new FileInputStream(file));
			
			byte[] buf = new byte[1024];
			int len;
			boolean done = false;
			boolean isAscii = true;
			
			while((len = bis.read(buf, 0, buf.length)) != -1)
			{
				if (isAscii) {isAscii = jchardetector.isAscii(buf,len);}
				if (!isAscii && !done) {done = jchardetector.DoIt(buf, len, false);}
			}
			
			jchardetector.DataEnd();
			
			if (isAscii) {return "ASCII";}
			else {return jchardetResult;}
		}
		catch (FileNotFoundException e) {e.printStackTrace(); return "";}
		catch (IOException e) {e.printStackTrace(); return "";}
	}
	
	/**
	 * An implementation of the {@see nsICharsetDetectionObserver} class.
	 */
	public void Notify(String string) {jchardetResult = string;}
	
	/**
	 * Returns the results of the {@see cpdetector.io.ParsingDetector}
	 * library.
	 * @param file - the file for character set detection
	 * @return String results of character set detection
	 */
	private static String getCpDetect(File file)
	{
		try
		{
			FileInputStream fis = new FileInputStream(file);
			ParsingDetector pd = new ParsingDetector();
			Charset charset = pd.detectCodepage(fis, (int) file.length());
			return charset.name();
		}
		catch (FileNotFoundException e) {e.printStackTrace(); return "";}
		catch (IOException e) {e.printStackTrace(); return "";}
	}
	
	/**
	 * Returns the results of the {@see com.ibm.icu.text.CharsetDetector}
	 * library.
	 * @param file - the file for character set detection
	 * @return String results of character set detection
	 */
	private static String getICU4J(File file)
	{
		try
		{
			CharsetDetector detector = new CharsetDetector();
			detector.setText(FileUtils.readBytes(file));
			return detector.detect().getName();
		}
		catch (FileNotFoundException e) {e.printStackTrace(); return "";}
		catch (IOException e) {e.printStackTrace(); return "";}
		
	}
	
	/**
	 * Returns the results of the {@see monq.stuff.EncodingDetector}
	 * library.
	 * @param file - the file for character set detection
	 * @return String results of character set detection
	 */
	private static String getMonq(File file)
	{
		try {return EncodingDetector.detect(new ByteArrayInputStream(FileUtils.readBytes(file)));}
		catch (FileNotFoundException e) {e.printStackTrace(); return "";}
		catch (IOException e) {e.printStackTrace(); return "";}
	}
	
	/**
	 * Returns the results of the {@see com.sun.syndication.io.XmlReader}
	 * library.
	 * @param file - the file for character set detection
	 * @return String results of character set detection
	 */
	private static String getRome(File file)
	{
		try {return new XmlReader(file).getEncoding();}
		catch (IOException e) {e.printStackTrace(); return "";}
	}
	
	/**
	 * Returns the results of the {@see com.glaforge.i18n.io.CharsetToolkit}
	 * library.
	 * @param file - the file for character set detection
	 * @return String results of character set detection
	 */
	private static String getGlaforge(File file)
	{
		try
		{
			return CharsetToolkit.guessEncoding(file, 4096).displayName();
		}
		catch (FileNotFoundException e) {e.printStackTrace(); return "";}
		catch (IOException e) {e.printStackTrace(); return "";}
	}
	
	/**
	 * Processes the arguments passed to the main method. Assigns values to
	 * the global variables:
	 * <code>testPath</code> and <code>outputFilename</code>.
	 * @param args
	 */
	private static void processArguments(String[] args)
	{
		if (args.length >= 1)
		{
			if ((args[0].equals("-help")) || (args[0].equals("-h"))) {usage();}
			else {testPath = args[0];}
		}
		if (args.length >= 2) {outputFilename = args[1];}
		
		/* default values if no parameters are passed */
		
		if (testPath == null) {testPath = SystemUtils.USER_DIR;}
		if (outputFilename == null) {outputFilename = "charset_detect_results.txt";}
	}
	
	/**
	 * Creates the output {@see java.io.FileWriter} to write the results of
	 * charset detection to a file. Writes the first line of headers to the file.
	 */
	private static void tryOutputFileWriter()
	{
		try
		{
			outputFileWriter =
				new FileWriter(
						new File(
								testPath + 
								SystemUtils.FILE_SEPARATOR + 
								outputFilename));
			
			outputFileWriter.write(
					"filename|" +
					"org.mozilla.intl.chardet.nsDetector|" +
					"cpdetector.io.ParsingDetector|" +
					"com.ibm.icu.text.CharsetDetector|" +
					"monq.stuff.EncodingDetector|" +
					"com.sun.syndication.io.XmlReader|" +
					"com.glaforge.i18n.io.CharsetToolkit\n");
		}
		catch (IOException e)
		{
			//e.printStackTrace();
			System.out.println("ERROR: Unable to create output file!");
			System.out.println(e.getLocalizedMessage());
			System.exit(1);
		}
	}
	
	/**
	 * Prints the copyright, license and uage paramaters.
	 */
	private static void usage()
	{
		System.out.println("");
		System.out.println("-------------------------------------------------");
		System.out.println("CharsetDetectorTester 0.1 build 20061228");
		System.out.println("Copyright (C) 2006 Fred Eaker <fredeaker@gmail.com>");
		System.out.println("-------------------------------------------------");
		System.out.println("This program is free software; you can redistribute it and/or");
		System.out.println("modify it under the terms of the GNU General Public License");
		System.out.println("as published by the Free Software Foundation; either version 2");
		System.out.println("of the License, or (at your option) any later version.");
		System.out.println("-------------------------------------------------");
		System.out.println("USAGE:");
		System.out.println("java CharsetDetectorTester.class [test directory] [output filename] ");
		System.out.println("");
		System.out.println("[test directory] should contain the files you would like to test.");
		System.out.println("");
		System.out.println("If no test directory option is given the current directory will be used.");
		System.out.println("");
		System.out.println("[output filename] is the name of the file which will contain");
		System.out.println("the results of testing. It will be located in the [test directory].");
		System.out.println("");
		System.exit(0);
	}
}